@@ -1,6 +1,6 @@
 /* Subroutines for insn-output.c for Windows NT.
    Contributed by Douglas Rupp (drupp@cs.washington.edu)
-   Copyright (C) 1995, 1997, 1998, 1999, 2000, 2001, 2002, 2003
+   Copyright (C) 1995, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004
    Free Software Foundation, Inc.
 
 This file is part of GCC.
@@ -22,6 +22,8 @@ Software Foundation, 59 Temple Place - Suite 330, Boston, MA
 
 #include "config.h"
 #include "system.h"
+#include "coretypes.h"
+#include "tm.h"
 #include "rtl.h"
 #include "regs.h"
 #include "hard-reg-set.h"
@@ -44,13 +46,12 @@ Software Foundation, 59 Temple Place - Suite 330, Boston, MA
    multiple times.
 */
 
-static tree associated_type PARAMS ((tree));
-const char * gen_stdcall_suffix PARAMS ((tree));
-const char * gen_fastcall_suffix PARAMS ((tree));
-int i386_pe_dllexport_p PARAMS ((tree));
-int i386_pe_dllimport_p PARAMS ((tree));
-void i386_pe_mark_dllexport PARAMS ((tree));
-void i386_pe_mark_dllimport PARAMS ((tree));
+static tree associated_type (tree);
+static tree gen_stdcall_or_fastcall_suffix (tree, bool);
+static int i386_pe_dllexport_p (tree);
+static int i386_pe_dllimport_p (tree);
+static void i386_pe_mark_dllexport (tree);
+static void i386_pe_mark_dllimport (tree);
 
 /* This is we how mark internal identifiers with dllimport or dllexport
    attributes.  */
@@ -61,19 +62,11 @@ void i386_pe_mark_dllimport PARAMS ((tree));
 #define DLL_EXPORT_PREFIX "#e."
 #endif
 
-#ifndef FASTCALL_PREFIX
-#define FASTCALL_PREFIX '@'
-#endif
-
 /* Handle a "dllimport" or "dllexport" attribute;
    arguments as in struct attribute_spec.handler.  */
 tree
-ix86_handle_dll_attribute (pnode, name, args, flags, no_add_attrs)
-     tree * pnode;
-     tree name;
-     tree args;
-     int flags;
-     bool *no_add_attrs;
+ix86_handle_dll_attribute (tree * pnode, tree name, tree args, int flags,
+			   bool *no_add_attrs)
 {
   tree node = *pnode;
 
@@ -107,7 +100,7 @@ ix86_handle_dll_attribute (pnode, name, args, flags, no_add_attrs)
       if (TREE_CODE (node) == FUNCTION_DECL  && DECL_INITIAL (node)
           && !DECL_INLINE (node))
 	{
-	  error_with_decl (node, "function `%s' definition is marked dllimport.");
+	  error ("%Jfunction `%D' definition is marked dllimport.", node, node);
 	  *no_add_attrs = true;
 	}
 
@@ -115,27 +108,28 @@ ix86_handle_dll_attribute (pnode, name, args, flags, no_add_attrs)
 	{
 	  if (DECL_INITIAL (node))
 	    {
-	      error_with_decl (node,"variable `%s' definition is marked dllimport.");
+	      error ("%Jvariable `%D' definition is marked dllimport.",
+		     node, node);
 	      *no_add_attrs = true;
 	    }
 
 	  /* `extern' needn't be specified with dllimport.
 	     Specify `extern' now and hope for the best.  Sigh.  */
-	  DECL_EXTERNAL (node) = 1; 
+	  DECL_EXTERNAL (node) = 1;
 	  /* Also, implicitly give dllimport'd variables declared within
 	     a function global scope, unless declared static.  */
 	  if (current_function_decl != NULL_TREE && !TREE_STATIC (node))
-  	    TREE_PUBLIC (node) = 1;
+	    TREE_PUBLIC (node) = 1;
 	}
     }
 
-  /*  Report error if symbol is not accessible at global scope. */
+  /*  Report error if symbol is not accessible at global scope.  */
   if (!TREE_PUBLIC (node)
       && (TREE_CODE (node) == VAR_DECL
-	  || TREE_CODE (node) == FUNCTION_DECL)) 
+	  || TREE_CODE (node) == FUNCTION_DECL))
     {
-      error_with_decl (node, "external linkage required for symbol '%s' because of '%s' attribute.",
-		       IDENTIFIER_POINTER (name));
+      error ("%Jexternal linkage required for symbol '%D' because of "
+	     "'%s' attribute.", node, node, IDENTIFIER_POINTER (name));
       *no_add_attrs = true;
     }
 
@@ -145,12 +139,9 @@ ix86_handle_dll_attribute (pnode, name, args, flags, no_add_attrs)
 /* Handle a "shared" attribute;
    arguments as in struct attribute_spec.handler.  */
 tree
-ix86_handle_shared_attribute (node, name, args, flags, no_add_attrs)
-     tree *node;
-     tree name;
-     tree args ATTRIBUTE_UNUSED;
-     int flags ATTRIBUTE_UNUSED;
-     bool *no_add_attrs;
+ix86_handle_shared_attribute (tree *node, tree name,
+			      tree args ATTRIBUTE_UNUSED,
+			      int flags ATTRIBUTE_UNUSED, bool *no_add_attrs)
 {
   if (TREE_CODE (*node) != VAR_DECL)
     {
@@ -166,8 +157,7 @@ ix86_handle_shared_attribute (node, name, args, flags, no_add_attrs)
    imported or exported.  */
 
 static tree
-associated_type (decl)
-     tree decl;
+associated_type (tree decl)
 {
   tree t = NULL_TREE;
 
@@ -180,7 +170,8 @@ associated_type (decl)
 	 dtor's are not affected by class status but virtual and
 	 non-virtual thunks are.  */
       if (!DECL_ARTIFICIAL (decl) || DECL_COMDAT (decl))
-	t = TREE_TYPE (TREE_VALUE (TYPE_ARG_TYPES (TREE_TYPE (decl))));
+	t = TYPE_MAIN_VARIANT
+	  (TREE_TYPE (TREE_VALUE (TYPE_ARG_TYPES (TREE_TYPE (decl)))));
     }
   else if (DECL_CONTEXT (decl)
 	   && TREE_CODE_CLASS (TREE_CODE (DECL_CONTEXT (decl))) == 't')
@@ -191,9 +182,8 @@ associated_type (decl)
 
 /* Return nonzero if DECL is a dllexport'd object.  */
 
-int
-i386_pe_dllexport_p (decl)
-     tree decl;
+static int
+i386_pe_dllexport_p (tree decl)
 {
   tree exp;
 
@@ -218,9 +208,8 @@ i386_pe_dllexport_p (decl)
 
 /* Return nonzero if DECL is a dllimport'd object.  */
 
-int
-i386_pe_dllimport_p (decl)
-     tree decl;
+static int
+i386_pe_dllimport_p (tree decl)
 {
   tree imp;
   int context_imp = 0;
@@ -255,29 +244,33 @@ i386_pe_dllimport_p (decl)
 	{
 	   /* Don't warn about artificial methods.  */
 	  if (!DECL_ARTIFICIAL (decl))
-	    warning_with_decl (decl,"function '%s' is defined after prior declaration as dllimport: attribute ignored.");
+	    warning ("%Jfunction '%D' is defined after prior declaration "
+		     "as dllimport: attribute ignored", decl, decl);
 	  return 0;
 	}
 
       /* We ignore the dllimport attribute for inline member functions.
 	 This differs from MSVC behavior which treats it like GNUC
-     	 'extern inline' extension.   */
+	 'extern inline' extension.  */
       else if (TREE_CODE (decl) == FUNCTION_DECL && DECL_INLINE (decl))
         {
 	  if (extra_warnings)
-	    warning_with_decl (decl, "inline function '%s' is declared as dllimport: attribute ignored.");
+	    warning ("%Jinline function '%D' is declared as dllimport: "
+		     "attribute ignored.", decl, decl);
 	  return 0;
 	}
 
       /*  Don't allow definitions of static data members in dllimport class,
-	  Just ignore attribute for vtable data.  */
+	  If vtable data is marked as DECL_EXTERNAL, import it; otherwise just
+          ignore the attribute.  */
       else if (TREE_CODE (decl) == VAR_DECL
 	       && TREE_STATIC (decl) && TREE_PUBLIC (decl)
 	       && !DECL_EXTERNAL (decl) && context_imp)
 	{
 	  if (!DECL_VIRTUAL_P (decl))
-	      error_with_decl (decl, "definition of static data member '%s' of dllimport'd class.");
-           return 0;
+            error ("%Jdefinition of static data member '%D' of "
+		   "dllimport'd class.", decl, decl);
+	  return 0;
 	}
 
       /* Since we can't treat a pointer to a dllimport'd symbol as a
@@ -298,8 +291,7 @@ i386_pe_dllimport_p (decl)
 /* Return nonzero if SYMBOL is marked as being dllexport'd.  */
 
 int
-i386_pe_dllexport_name_p (symbol)
-     const char *symbol;
+i386_pe_dllexport_name_p (const char *symbol)
 {
   return (strncmp (DLL_EXPORT_PREFIX, symbol,
 		   strlen (DLL_EXPORT_PREFIX)) == 0);
@@ -308,8 +300,7 @@ i386_pe_dllexport_name_p (symbol)
 /* Return nonzero if SYMBOL is marked as being dllimport'd.  */
 
 int
-i386_pe_dllimport_name_p (symbol)
-     const char *symbol;
+i386_pe_dllimport_name_p (const char *symbol)
 {
   return (strncmp (DLL_IMPORT_PREFIX, symbol,
 		   strlen (DLL_IMPORT_PREFIX)) == 0);
@@ -318,9 +309,8 @@ i386_pe_dllimport_name_p (symbol)
 /* Mark a DECL as being dllexport'd.
    Note that we override the previous setting (eg: dllimport).  */
 
-void
-i386_pe_mark_dllexport (decl)
-     tree decl;
+static void
+i386_pe_mark_dllexport (tree decl)
 {
   const char *oldname;
   char  *newname;
@@ -337,7 +327,8 @@ i386_pe_mark_dllexport (decl)
     abort ();
   if (i386_pe_dllimport_name_p (oldname))
     {
-      warning_with_decl (decl,"inconsistent dll linkage for '%s': dllexport assumed.");
+      warning ("%Jinconsistent dll linkage for '%D', dllexport assumed.",
+	       decl, decl);
      /* Remove DLL_IMPORT_PREFIX.  */
       oldname += strlen (DLL_IMPORT_PREFIX);
       DECL_NON_ADDR_CONST_P (decl) = 0;
@@ -360,9 +351,8 @@ i386_pe_mark_dllexport (decl)
 
 /* Mark a DECL as being dllimport'd.  */
 
-void
-i386_pe_mark_dllimport (decl)
-     tree decl;
+static void
+i386_pe_mark_dllimport (tree decl)
 {
   const char *oldname;
   char  *newname;
@@ -385,10 +375,11 @@ i386_pe_mark_dllimport (decl)
     }
   else if (i386_pe_dllimport_name_p (oldname))
     {
-      /* Already done, but do a sanity check to prevent assembler errors. */
+      /* Already done, but do a sanity check to prevent assembler errors.  */
       if (!DECL_EXTERNAL (decl) || !TREE_PUBLIC (decl))
 	{
-	  error_with_decl (decl, "failure in redeclaration of '%s': dllimport'd symbol lacks external linkage.");
+	  error ("%Jfailure in redeclaration of '%D': dllimport'd "
+		 "symbol lacks external linkage.", decl, decl);
 	  abort();
 	}
       return;
@@ -412,107 +403,87 @@ i386_pe_mark_dllimport (decl)
   DECL_NON_ADDR_CONST_P (decl) = 1;
 }
 
-/* Return string which is the former assembler name modified with a 
-   prefix consisting of FASTCALL_PREFIX and a suffix consisting of an
-   atsign (@) followed by the number of bytes of arguments.  */
+/* Return string which is the former assembler name modified with a
+   suffix consisting of an atsign (@) followed by the number of bytes of
+   arguments.  If FASTCALL is true, also add the FASTCALL_PREFIX.  */
 
-const char *
-gen_fastcall_suffix (decl)
-  tree decl;
-{
-  int total = 0;
-
-  const char *asmname = IDENTIFIER_POINTER (DECL_ASSEMBLER_NAME (decl));
-  char *newsym;
-
-  if (TYPE_ARG_TYPES (TREE_TYPE (decl)))
-    if (TREE_VALUE (tree_last (TYPE_ARG_TYPES (TREE_TYPE (decl))))
-        == void_type_node)
-      {
-	tree formal_type = TYPE_ARG_TYPES (TREE_TYPE (decl));
-
-	while (TREE_VALUE (formal_type) != void_type_node)
-	  {
-	    int parm_size
-	      = TREE_INT_CST_LOW (TYPE_SIZE (TREE_VALUE (formal_type)));
-	    /* Must round up to include padding.  This is done the same
-	       way as in store_one_arg.  */
-	    parm_size = ((parm_size + PARM_BOUNDARY - 1)
-			 / PARM_BOUNDARY * PARM_BOUNDARY);
-	    total += parm_size;
-	    formal_type = TREE_CHAIN (formal_type);
-	  }
-      }
-
-  /* Assume max of 8 base 10 digits in the suffix.  */ 
-  newsym = xmalloc (1 + strlen (asmname) + 1 + 8 + 1);
-  sprintf (newsym, "%c%s@%d", FASTCALL_PREFIX, asmname, total/BITS_PER_UNIT);
-  return IDENTIFIER_POINTER (get_identifier (newsym));
-}
-
-/* Return string which is the former assembler name modified with a 
-   suffix consisting of an atsign (@) followed by the number of bytes of 
-   arguments */
-
-const char *
-gen_stdcall_suffix (decl)
-  tree decl;
+static tree
+gen_stdcall_or_fastcall_suffix (tree decl, bool fastcall)
 {
   int total = 0;
   /* ??? This probably should use XSTR (XEXP (DECL_RTL (decl), 0), 0) instead
      of DECL_ASSEMBLER_NAME.  */
   const char *asmname = IDENTIFIER_POINTER (DECL_ASSEMBLER_NAME (decl));
   char *newsym;
+  char *p;
+  tree formal_type;
 
-  if (TYPE_ARG_TYPES (TREE_TYPE (decl)))
-    if (TREE_VALUE (tree_last (TYPE_ARG_TYPES (TREE_TYPE (decl)))) 
-        == void_type_node)
-      {
-	tree formal_type = TYPE_ARG_TYPES (TREE_TYPE (decl));
+  /* Do not change the identifier if a verbatim asmspec or already done. */
+  if (*asmname == '*' || strchr (asmname, '@'))
+    return DECL_ASSEMBLER_NAME (decl);
 
-	while (TREE_VALUE (formal_type) != void_type_node)
-	  {
-	    int parm_size
-	      = TREE_INT_CST_LOW (TYPE_SIZE (TREE_VALUE (formal_type)));
+  formal_type = TYPE_ARG_TYPES (TREE_TYPE (decl));
+  if (formal_type != NULL_TREE)
+    {
+      /* These attributes are ignored for variadic functions in
+	 i386.c:ix86_return_pops_args. For compatibility with MS
+         compiler do not add @0 suffix here.  */ 
+      if (TREE_VALUE (tree_last (formal_type)) != void_type_node)
+        return DECL_ASSEMBLER_NAME (decl);
+
+      /* Quit if we hit an incomplete type.  Error is reported
+         by convert_arguments in c-typeck.c or cp/typeck.c.  */
+      while (TREE_VALUE (formal_type) != void_type_node
+	     && COMPLETE_TYPE_P (TREE_VALUE (formal_type)))	
+	{
+	  int parm_size
+	    = TREE_INT_CST_LOW (TYPE_SIZE (TREE_VALUE (formal_type)));
 	    /* Must round up to include padding.  This is done the same
 	       way as in store_one_arg.  */
-	    parm_size = ((parm_size + PARM_BOUNDARY - 1)
-			 / PARM_BOUNDARY * PARM_BOUNDARY);
-	    total += parm_size;
-	    formal_type = TREE_CHAIN (formal_type);
-	  }
-      }
+	  parm_size = ((parm_size + PARM_BOUNDARY - 1)
+		       / PARM_BOUNDARY * PARM_BOUNDARY);
+	  total += parm_size;
+	  formal_type = TREE_CHAIN (formal_type);\
+	}
+     }
 
-  /* Assume max of 8 base 10 digits in the suffix.  */ 
-  newsym = xmalloc (strlen (asmname) + 1 + 8 + 1);
-  sprintf (newsym, "%s@%d", asmname, total/BITS_PER_UNIT);
-  return IDENTIFIER_POINTER (get_identifier (newsym));
+  /* Assume max of 8 base 10 digits in the suffix.  */
+  newsym = alloca (1 + strlen (asmname) + 1 + 8 + 1);
+  p = newsym;
+  if (fastcall)
+    *p++ = FASTCALL_PREFIX;
+  sprintf (p, "%s@%d", asmname, total/BITS_PER_UNIT);
+  return get_identifier (newsym);
 }
 
 void
-i386_pe_encode_section_info (decl, first)
-     tree decl;
-     int first ATTRIBUTE_UNUSED;
+i386_pe_encode_section_info (tree decl, rtx rtl, int first)
 {
-   /* This bit is copied from i386.h.  */
-  if (optimize > 0 && TREE_CONSTANT (decl)
-      && (!flag_writable_strings || TREE_CODE (decl) != STRING_CST))
-    {
-      rtx rtl = (TREE_CODE_CLASS (TREE_CODE (decl)) != 'd'
-                 ? TREE_CST_RTL (decl) : DECL_RTL (decl));
-      SYMBOL_REF_FLAG (XEXP (rtl, 0)) = 1;
-    }
+  default_encode_section_info (decl, rtl, first);
 
-  if (TREE_CODE (decl) == FUNCTION_DECL)
+  if (first && TREE_CODE (decl) == FUNCTION_DECL)
     {
-      if (lookup_attribute ("stdcall",
-			    TYPE_ATTRIBUTES (TREE_TYPE (decl))))
-        XEXP (DECL_RTL (decl), 0) = 
-	  gen_rtx (SYMBOL_REF, Pmode, gen_stdcall_suffix (decl));
-      else if (lookup_attribute ("fastcall",
-				 TYPE_ATTRIBUTES (TREE_TYPE (decl))))
-        XEXP (DECL_RTL (decl), 0) =
-	  gen_rtx (SYMBOL_REF, Pmode, gen_fastcall_suffix (decl));
+      tree type_attributes = TYPE_ATTRIBUTES (TREE_TYPE (decl));
+      tree newid = NULL_TREE;
+
+      if (lookup_attribute ("stdcall", type_attributes))
+	newid = gen_stdcall_or_fastcall_suffix (decl, false);
+      else if (lookup_attribute ("fastcall", type_attributes))
+	newid = gen_stdcall_or_fastcall_suffix (decl, true);
+      if (newid != NULL_TREE) 	
+	{
+	  rtx rtlname = XEXP (rtl, 0);
+	  if (GET_CODE (rtlname) == MEM)
+	    rtlname = XEXP (rtlname, 0);
+	  XSTR (rtlname, 0) = IDENTIFIER_POINTER (newid);
+#if 0
+	  /* These attributes must be present on first declaration,
+	     change_decl_assembler_name will warn if they are added
+	     later and the decl has been referenced, but duplicate_decls
+	     should catch the mismatch before this is called.  */ 
+          change_decl_assembler_name (decl, newid);
+#endif
+	}
     }
 
   /* Mark the decl so we can tell from the rtl whether the object is
@@ -536,13 +507,17 @@ i386_pe_encode_section_info (decl, first)
 	   && i386_pe_dllimport_name_p (XSTR (XEXP (XEXP (DECL_RTL (decl), 0), 0), 0)))
     {
       const char *oldname = XSTR (XEXP (XEXP (DECL_RTL (decl), 0), 0), 0);
+
       /* Remove DLL_IMPORT_PREFIX.  */
       tree idp = get_identifier (oldname + strlen (DLL_IMPORT_PREFIX));
       rtx newrtl = gen_rtx (SYMBOL_REF, Pmode, IDENTIFIER_POINTER (idp));
 
-      warning_with_decl (decl, "'%s' %s after being referenced with dllimport linkage.",
-	         	 (DECL_INITIAL (decl) || !DECL_EXTERNAL (decl))
-			 ? "defined locally" : "redeclared without dllimport attribute");
+      if (DECL_INITIAL (decl) || !DECL_EXTERNAL (decl))
+	warning ("%J'%D' defined locally after being "
+		 "referenced with dllimport linkage", decl, decl);
+      else
+	warning ("%J'%D' redeclared without dllimport attribute "
+		 "after being referenced with dllimport linkage", decl, decl);
 
       XEXP (DECL_RTL (decl), 0) = newrtl;
 
@@ -557,8 +532,7 @@ i386_pe_encode_section_info (decl, first)
    prefix if it exists.  */
 
 const char *
-i386_pe_strip_name_encoding (str)
-     const char *str;
+i386_pe_strip_name_encoding (const char *str)
 {
   if (strncmp (str, DLL_IMPORT_PREFIX, strlen (DLL_IMPORT_PREFIX))
       == 0)
@@ -571,15 +545,19 @@ i386_pe_strip_name_encoding (str)
   return str;
 }
 
-/* Also strip the stdcall suffix.  */
+/* Also strip the fastcall prefix and stdcall suffix.  */
 
 const char *
-i386_pe_strip_name_encoding_full (str)
-     const char *str;
+i386_pe_strip_name_encoding_full (const char *str)
 {
   const char *p;
   const char *name = i386_pe_strip_name_encoding (str);
- 
+
+  /* Strip leading '@' on fastcall symbols.  */
+  if (*name == '@')
+    name++;
+
+  /* Strip trailing '@n'.  */
   p = strchr (name, '@');
   if (p)
     return ggc_alloc_string (name, p - name);
@@ -591,31 +569,29 @@ i386_pe_strip_name_encoding_full (str)
    whereas symbols for functions using other calling conventions don't
    have a prefix (unless they are marked dllimport or dllexport).  */
 
-void i386_pe_output_labelref (stream, name)
-     FILE *stream;
-     const char *name;
+void i386_pe_output_labelref (FILE *stream, const char *name)
 {
   if (strncmp (name, DLL_IMPORT_PREFIX, strlen (DLL_IMPORT_PREFIX))
       == 0)
-    /* A dll import */ 
+    /* A dll import */
     {
       if (name[strlen (DLL_IMPORT_PREFIX)] == FASTCALL_PREFIX)
-      /* A dllimport fastcall symbol.  */   
+      /* A dllimport fastcall symbol.  */
         {
           fprintf (stream, "__imp_%s",
                    i386_pe_strip_name_encoding (name));
         }
       else
-      /* A dllimport non-fastcall symbol.  */ 
+      /* A dllimport non-fastcall symbol.  */
         {
           fprintf (stream, "__imp__%s",
                    i386_pe_strip_name_encoding (name));
         }
     }
   else if ((name[0] == FASTCALL_PREFIX)
-           || (strncmp (name, DLL_EXPORT_PREFIX, strlen (DLL_EXPORT_PREFIX)
-	       == 0 
-	       && name[strlen (DLL_EXPORT_PREFIX)] == FASTCALL_PREFIX)))
+           || (strncmp (name, DLL_EXPORT_PREFIX, strlen (DLL_EXPORT_PREFIX))
+	       == 0
+	       && name[strlen (DLL_EXPORT_PREFIX)] == FASTCALL_PREFIX))
     /* A fastcall symbol.  */
     {
       fprintf (stream, "%s",
@@ -630,9 +606,7 @@ void i386_pe_output_labelref (stream, name)
 }
 
 void
-i386_pe_unique_section (decl, reloc)
-     tree decl;
-     int reloc;
+i386_pe_unique_section (tree decl, int reloc)
 {
   int len;
   const char *name, *prefix;
@@ -644,7 +618,7 @@ i386_pe_unique_section (decl, reloc)
   /* The object is put in, for example, section .text$foo.
      The linker will then ultimately place them in .text
      (everything from the $ on is stripped). Don't put
-     read-only data in .rdata section to avoid a PE linker 
+     read-only data in .rdata section to avoid a PE linker
      bug when .rdata$* grouped sections are used in code
      without a .rdata section.  */
   if (TREE_CODE (decl) == FUNCTION_DECL)
@@ -678,10 +652,7 @@ i386_pe_unique_section (decl, reloc)
 #define SECTION_PE_SHARED	SECTION_MACH_DEP
 
 unsigned int
-i386_pe_section_type_flags (decl, name, reloc)
-     tree decl;
-     const char *name;
-     int reloc;
+i386_pe_section_type_flags (tree decl, const char *name, int reloc)
 {
   static htab_t htab;
   unsigned int flags;
@@ -719,25 +690,33 @@ i386_pe_section_type_flags (decl, name, reloc)
   else
     {
       if (decl && **slot != flags)
-	error_with_decl (decl, "%s causes a section type conflict");
+	error ("%J'%D' causes a section type conflict", decl, decl);
     }
 
   return flags;
 }
 
 void
-i386_pe_asm_named_section (name, flags)
-     const char *name;
-     unsigned int flags;
+i386_pe_asm_named_section (const char *name, unsigned int flags)
 {
   char flagchars[8], *f = flagchars;
 
-  if (flags & SECTION_CODE)
-    *f++ = 'x';
-  if (flags & SECTION_WRITE)
-    *f++ = 'w';
-  if (flags & SECTION_PE_SHARED)
-    *f++ = 's';
+  if ((flags & (SECTION_CODE | SECTION_WRITE)) == 0)
+    /* readonly data */
+    {
+      *f++ ='d';  /* This is necessary for older versions of gas.  */
+      *f++ ='r';
+    }
+  else	
+    {
+      if (flags & SECTION_CODE)
+        *f++ = 'x';
+      if (flags & SECTION_WRITE)
+        *f++ = 'w';
+      if (flags & SECTION_PE_SHARED)
+        *f++ = 's';
+    }
+
   *f = '\0';
 
   fprintf (asm_out_file, "\t.section\t%s,\"%s\"\n", name, flagchars);
@@ -765,10 +744,7 @@ i386_pe_asm_named_section (name, flags)
    visible.  */
 
 void
-i386_pe_declare_function_type (file, name, public)
-     FILE *file;
-     const char *name;
-     int public;
+i386_pe_declare_function_type (FILE *file, const char *name, int public)
 {
   fprintf (file, "\t.def\t");
   assemble_name (file, name);
@@ -779,13 +755,13 @@ i386_pe_declare_function_type (file, name, public)
 
 /* Keep a list of external functions.  */
 
-struct extern_list
+struct extern_list GTY(())
 {
   struct extern_list *next;
   const char *name;
 };
 
-static struct extern_list *extern_head;
+static GTY(()) struct extern_list *extern_head;
 
 /* Assemble an external function reference.  We need to keep a list of
    these, so that we can output the function types at the end of the
@@ -794,12 +770,11 @@ static struct extern_list *extern_head;
    for it then.  */
 
 void
-i386_pe_record_external_function (name)
-     const char *name;
+i386_pe_record_external_function (const char *name)
 {
   struct extern_list *p;
 
-  p = (struct extern_list *) xmalloc (sizeof *p);
+  p = (struct extern_list *) ggc_alloc (sizeof *p);
   p->next = extern_head;
   p->name = name;
   extern_head = p;
@@ -807,29 +782,27 @@ i386_pe_record_external_function (name)
 
 /* Keep a list of exported symbols.  */
 
-struct export_list
+struct export_list GTY(())
 {
   struct export_list *next;
   const char *name;
   int is_data;		/* used to type tag exported symbols.  */
 };
 
-static struct export_list *export_head;
+static GTY(()) struct export_list *export_head;
 
 /* Assemble an export symbol entry.  We need to keep a list of
    these, so that we can output the export list at the end of the
    assembly.  We used to output these export symbols in each function,
-   but that causes problems with GNU ld when the sections are 
+   but that causes problems with GNU ld when the sections are
    linkonce.  */
 
 void
-i386_pe_record_exported_symbol (name, is_data)
-     const char *name;
-     int is_data;
+i386_pe_record_exported_symbol (const char *name, int is_data)
 {
   struct export_list *p;
 
-  p = (struct export_list *) xmalloc (sizeof *p);
+  p = (struct export_list *) ggc_alloc (sizeof *p);
   p->next = export_head;
   p->name = name;
   p->is_data = is_data;
@@ -841,27 +814,26 @@ i386_pe_record_exported_symbol (name, is_data)
    output the .drectve section.  */
 
 void
-i386_pe_asm_file_end (file)
-     FILE *file;
+i386_pe_file_end (void)
 {
   struct extern_list *p;
 
-  ix86_asm_file_end (file);
+  ix86_file_end ();
+  if (write_symbols != SDB_DEBUG)
+    for (p = extern_head; p != NULL; p = p->next)
+      {
+	tree decl;
 
-  for (p = extern_head; p != NULL; p = p->next)
-    {
-      tree decl;
+	decl = get_identifier (p->name);
 
-      decl = get_identifier (p->name);
-
-      /* Positively ensure only one declaration for any given symbol.  */
-      if (! TREE_ASM_WRITTEN (decl) && TREE_SYMBOL_REFERENCED (decl))
-	{
-	  TREE_ASM_WRITTEN (decl) = 1;
-	  i386_pe_declare_function_type (asm_out_file, p->name,
-			  		 TREE_PUBLIC (decl));
-	}
-    }
+	/* Positively ensure only one declaration for any given symbol.  */
+	if (! TREE_ASM_WRITTEN (decl) && TREE_SYMBOL_REFERENCED (decl))
+	  {
+	    TREE_ASM_WRITTEN (decl) = 1;
+	    i386_pe_declare_function_type (asm_out_file, p->name,
+					   TREE_PUBLIC (decl));
+	  }
+      }
 
   if (export_head)
     {
@@ -875,3 +847,5 @@ i386_pe_asm_file_end (file)
 	}
     }
 }
+
+#include "gt-winnt.h"
