@@ -1,4 +1,3 @@
-#include "ansidecl.h"
 #include "version.h"
 
 /* This is the string reported as the version number by all components
@@ -6,7 +5,7 @@
    please modify this string to indicate that, e.g. by putting your
    organization's name in parentheses at the end of the string.  */
 
-const char version_string[] = "3.3.1 (mingw special 20030804-1)";
+const char version_string[] = "3.4.4 (mingw special)";
 
 /* This is the location of the online document giving instructions for
    reporting bugs.  If you distribute a modified version of GCC,
@@ -16,3 +15,4 @@ const char version_string[] = "3.3.1 (mingw special 20030804-1)";
    not bugs in your modifications.)  */
 
 const char bug_report_url[] = "<URL:http://www.mingw.org/bugs.shtml>";
+
